class RequestException(Exception):
    """Exception representing an error due to an incorrect request structure
    or missing required fields."""

    def __init__(self, status_code, message):
        Exception.__init__(self)
        self._status_code = status_code
        self._message = message

    def __str__(self):
        return "%s (HTTP Status: %s)" % (self._message, self._status_code)


class UnauthorizedException(RequestException):
    """Exception for unauthenticated or unauthorized request."""
    pass


class InvalidInputException(Exception):
    """Exception representing invalid input passed to the API Client."""
    pass



"""
This module provides an authentication implementation for requests to the HouseCanary API.
"""

import hmac
import hashlib
import datetime
import calendar

try:
    # Python 3
    from urllib.parse import urlencode, urlparse, parse_qsl, urlunparse
except ImportError:
    # Python 2
    from urllib import urlencode
    from urlparse import urlparse, parse_qsl, urlunparse

import requests


class HCAuth(requests.auth.AuthBase):
    """ HouseCanary Authentication handler for Requests."""

    def __init__(self, auth_key, auth_secret):
        """Create an authentication handler for HouseCanary API V1 requests

        Args:
            auth_key (string) - The HouseCanary API auth key
            auth_secret (string) - The HouseCanary API secret
        """

        self._auth_key = auth_key
        self._auth_secret = auth_secret

    def __call__(self, request):
        # Override in subclass
        return request


class HCAuthV1(HCAuth):
    """ HouseCanary API V1 Authentication handler for Requests.

    Generates an HMAC-SHA1 signature for the X-Auth-Signature header.
    """

    SIGN_DELIMITER = "\n"
    AUTH_PROTO_V1 = "hc_hmac_v1"

    def __call__(self, request):
        # parse the url of the request
        scheme, netloc, path, params, query, fragment = urlparse(request.url)

        # convert the query string to a dict
        query_params = dict(parse_qsl(query))

        # add additional query params that are required for authentication
        query_params['AuthKey'] = self._auth_key
        query_params['AuthProto'] = self.AUTH_PROTO_V1
        query_params['AuthTimestamp'] = str(self._get_timestamp_utc())

        # back to string
        query_string = urlencode(query_params, True)

        # get the post data from the request, if any
        data = request.body or ""

        # set the auth signature as required by the HouseCanary API
        signature = self._get_signature(path, query_string, request.method, data)
        request.headers["X-Auth-Signature"] = signature

        # recreate the url with the updated query string
        updated_url = urlunparse([scheme, netloc, path, params, query_string, fragment])
        request.url = requests.utils.requote_uri(updated_url)

        return request

    def _get_signature(self, endpoint_path, query_string, http_method, data):
        # all the inputs should be strings

        # uses hmac to produce a signature from the api secret and a message.
        # The message is defined as:
        # (HTTP_METHOD, HTTP_LOCATION, HTTP_QUERY_STRING, HTTP_POST_BODY)
        # concatenated by newline "\n" (unix style).
        sign_str = self.SIGN_DELIMITER.join([http_method, endpoint_path, query_string, data])
        signature = hmac.new(str(self._auth_secret), sign_str, digestmod=hashlib.sha1).hexdigest()
        return signature

    @staticmethod
    def _get_timestamp_utc():
        cur_utc_time = calendar.timegm(datetime.datetime.utcnow().utctimetuple())
        return cur_utc_time



"""
Provides Response to encapsulate API responses.
"""

from housecanary.object import Property


class Response(object):
    """Encapsulate an API reponse."""

    def __init__(self, endpoint_name, json_body, original_response):
        """
        Args:
            endpoint_name (str) - The endpoint of the request, such as "property/value"
            json_body - The response body in json format.
            original_response (response object) - server response returned from an http request.
        """
        self._endpoint_name = endpoint_name
        self._json_body = json_body
        self._response = original_response
        self._objects = []
        self._has_object_error = None
        self._object_errors = None

    @classmethod
    def create(cls, endpoint_name, json_body, original_response):
        """Factory for creating the correct type of Response based on the data.
        Args:
            endpoint_name (str) - The endpoint of the request, such as "property/value"
            json_body - The response body in json format.
            original_response (response object) - server response returned from an http request.
        """

        if endpoint_name == "property/value_report":
            return ValueReportResponse(endpoint_name, json_body, original_response)
        else:
            return PropertyResponse(endpoint_name, json_body, original_response)

    @property
    def endpoint_name(self):
        """Get the component name of the original request.

        Returns:
            Component name as a string.
        """
        return self._endpoint_name

    @property
    def response(self):
        """Gets the original response

        Returns:
            response object passed in during instantiation.
        """
        return self._response

    def json(self):
        """Gets the response body as json

        Returns:
            Json of the response body
        """
        return self._json_body

    def get_object_errors(self):
        """Gets a list of business error message strings
        for each of the requested objects that had a business error.
        If there was no error, returns an empty list

        Returns:
            List of strings
        """
        if self._object_errors is None:
            self._object_errors = [{str(o): o.get_errors()}
                                   for o in self.objects()
                                   if o.has_error()]

        return self._object_errors

    def has_object_error(self):
        """Returns true if any requested object had a business logic error,
        otherwise returns false

        Returns:
            boolean
        """
        if self._has_object_error is None:
            # scan the objects for any business error codes
            self._has_object_error = next(
                (True for o in self.objects()
                 if o.has_error()),
                False)
        return self._has_object_error

    def objects(self):
        """Override in subclasses"""
        raise NotImplementedError()


class PropertyResponse(Response):
    """Represents a single property and its data returned from the API."""
    def objects(self):
        """Gets a list of Property objects for the requested properties,
        each containing the property's returned json data from the API.

        Returns an empty list if the request format was PDF.

        Returns:
            List of Property objects
        """
        if not self._objects:
            body = self.json()

            self._objects = []

            if not isinstance(body, list):
                # The property endpoints return a list in the body.
                # This could maybe raise an exception.
                return []

            for address in body:
                prop = Property.create_from_json(address)
                self._objects.append(prop)

        return self._objects

    def properties(self):
        """Alias method for objects."""
        return self.objects()


class ZipcodeResponse(Response):
    """To be implemented later."""
    pass


class LatLngResponse(Response):
    """To be implemented later."""
    pass


class ValueReportResponse(Response):
    """The response from a value_report request."""

    def objects(self):
        """The value_report endpoint returns a json dict
           instead of a list of address results."""
        return []



"""
Provides a base client for making API requests.
"""

import requests


class RequestClient(object):
    """Base class for making http requests with the 'requests' lib."""

    def __init__(self, output_generator=None, authenticator=None):
        """
        Args:
            output_generator - Optional. An instance of an OutputGenerator that implements
                               a `process_response` method. Can be used to implement custom
                               serialization of the response data from the `execute_request`
                               method. If not specified, `execute_request` returns the
                               response from the requests lib unchanged.
            authenticator - Optional. An instance of a requests.auth.AuthBase implementation
                            for providing authentication to the request.
        """
        self._output_generator = output_generator
        self._auth = authenticator

    def execute_request(self, url, http_method, query_params, post_data):
        """Makes a request to the specified url endpoint with the
        specified http method, params and post data.

        Args:
            url (string): The url to the API without query params.
                          Example: "https://api.housecanary.com/v2/property/value"
            http_method (string): The http method to use for the request.
            query_params (dict): Dictionary of query params to add to the request.
            post_data: Json post data to send in the body of the request.

        Returns:
            The result of calling this instance's OutputGenerator process_response method
            on the requests.Response object.
            If no OutputGenerator is specified for this instance, returns the requests.Response.
        """

        response = requests.request(http_method, url, params=query_params,
                                    auth=self._auth, json=post_data)

        if isinstance(self._output_generator, str) and self._output_generator.lower() == "json":
            # shortcut for just getting json back
            return response.json()
        elif self._output_generator is not None:
            return self._output_generator.process_response(response)
        else:
            return response

    def get(self, url, query_params):
        """Makes a GET request to the specified url endpoint.

        Args:
            url (string): The url to the API without query params.
                          Example: "https://api.housecanary.com/v2/property/value"
            query_params (dict): Dictionary of query params to add to the request.

        Returns:
            The result of calling this instance's OutputGenerator process_response method
            on the requests.Response object.
            If no OutputGenerator is specified for this instance, returns the requests.Response.
        """
        return self.execute_request(url, "GET", query_params, None)

    def post(self, url, post_data, query_params=None):
        """Makes a POST request to the specified url endpoint.

        Args:
            url (string): The url to the API without query params.
                          Example: "https://api.housecanary.com/v2/property/value"
            post_data: Json post data to send in the body of the request.
            query_params (dict): Optional. Dictionary of query params to add to the request.

        Returns:
            The result of calling this instance's OutputGenerator process_response method
            on the requests.Response object.
            If no OutputGenerator is specified for this instance, returns the requests.Response.
        """
        if query_params is None:
            query_params = {}

        return self.execute_request(url, "POST", query_params, post_data)



"""
Provides a HouseCanaryObject class which encapsulates
an object and its associated data.
Currently, only the Property subclass is implemented.
"""

import housecanary.constants as constants


class HouseCanaryObject(object):
    """Base class for returned API objects."""

    def __init__(self):
        """
        Args:
            data - Json data returned from the API for this object.
            api_code - The HouseCanary business logic error code.
            api_code_description - The HouseCanary business logic error description.
        """
        self.component_results = []

    def has_error(self):
        """Returns whether there was a business logic error when fetching data
        for any components for this property.

        Returns:
            boolean
        """
        return next(
            (True for cr in self.component_results
             if cr.has_error()),
            False
        )

    def get_errors(self):
        """If there were any business errors fetching data for this property,
        returns the error messages.

        Returns:
            string - the error message, or None if there was no error.

        """
        return [{cr.component_name: cr.get_error()}
                for cr in self.component_results if cr.has_error()]

    def __str__(self):
        return "HouseCanaryObject"


class Property(HouseCanaryObject):
    """A single address"""

    def __init__(self, address=None, zipcode=None):
        """
        Args:
            address (required) -- Building number, street name and unit number.
            zipcode (required) -- Zipcode that matches the address.
            data (optional) -- The data returned from the API for this property.
            api_code (optional) -- The HouseCanary business logic
                error code reflecting any error with this property.
            api_code_description (optional) -- The HouseCanary business logic
                error description.
        """
        super(Property, self).__init__()

        self.address = str(address)
        self.zipcode = str(zipcode)
        self.zipcode_plus4 = None
        self.address_full = None
        self.city = None
        self.county_fips = None
        self.lat = None
        self.lng = None
        self.state = None
        self.unit = None
        self.meta = None

    @classmethod
    def create_from_json(cls, json_data):
        """Deserialize property json data into a Property object

        Args:
            json_data (dict): The json data for this property

        Returns:
            Property object

        """
        prop = Property()
        address_info = json_data["address_info"]
        prop.address = address_info["address"]
        prop.zipcode = address_info["zipcode"]
        prop.zipcode_plus4 = address_info["zipcode_plus4"]
        prop.address_full = address_info["address_full"]
        prop.city = address_info["city"]
        prop.county_fips = address_info["county_fips"]
        prop.lat = address_info["lat"]
        prop.lng = address_info["lng"]
        prop.state = address_info["state"]
        prop.unit = address_info["unit"]

        prop.meta = None
        if "meta" in json_data:
            prop.meta = json_data["meta"]

        # create a ComponentResult from each returned component's data
        # and append to this property's component_results list.
        for key, value in json_data.iteritems():
            if key not in ["address_info", "meta"]:
                component_result = ComponentResult(
                    key,
                    value["result"],
                    value["api_code"],
                    value["api_code_description"]
                )

                prop.component_results.append(component_result)

        return prop

    def __str__(self):
        return self.address


class Zipcode(HouseCanaryObject):
    pass


class LatLng(HouseCanaryObject):
    pass


class ComponentResult(object):
    """The results of a single component"""

    def __init__(self, component_name, json_data, api_code, api_code_description):
        """
        Args:
            component_name - string name of the component.
            json_data - Json data returned from the API for this object.
            api_code - The HouseCanary business logic error code.
            api_code_description - The HouseCanary business logic error description.
        """
        self.component_name = component_name
        self.json_data = json_data
        self.api_code = api_code
        self.api_code_description = api_code_description

    def has_error(self):
        """Returns whether this component had a business logic error"""
        return self.api_code > constants.BIZ_CODE_OK

    def get_error(self):
        """Gets the error of this component, if any"""
        return self.api_code_description



"""
This module provides a base class and two implementations of OutputGenerator,
which is responsible for processing a response before returning a result to the caller.

An OutputGenerator must implement a `process_response` method
which takes an input (usually a server's response to an HTTP request)
and returns something useful to the caller. This is a good place to do
custom serialization of an API response.
"""

try:
    # Python 3
    from urllib.parse import urlparse
except ImportError:
    # Python 2
    from urlparse import urlparse

from housecanary.response import Response
import housecanary.exceptions
import housecanary.constants as constants


class OutputGenerator(object):
    """Base class of an OutputGenerator. This base class just returns the given response."""

    def process_response(self, response):
        """Simply returns the response passed in."""
        return response


class JsonOutputGenerator(OutputGenerator):
    """Returns the JSON body of the response

    Expects the given response to implement a `json` method,
    like the response from the requests library.
    """

    def process_response(self, response):
        """Return the result of calling the json method on response"""
        return response.json()


class ResponseOutputGenerator(OutputGenerator):
    """Serializes the response into an instance of Response.

    Expects the given response to be a response from the requests library.
    """

    def process_response(self, response):
        content_type = response.headers['content-type']

        if content_type == "application/json":
            return self.process_json_response(response)
        elif content_type == "application/pdf":
            return self.process_pdf_response(response)
        elif content_type == "appliction/zip":
            return self.process_zip_response(response)
        else:
            return response

    def process_json_response(self, response):
        """For a json response, check if there was any error and throw exception.
        Otherwise, create a housecanary.response.Response."""

        response_json = response.json()

        # handle errors
        code_key = "code"
        if code_key in response_json and response_json[code_key] != constants.HTTP_CODE_OK:
            code = response_json[code_key]

            message = response_json
            if "message" in response_json:
                message = response_json["message"]
            elif "code_description" in response_json:
                message = response_json["code_description"]

            if code == constants.HTTP_FORBIDDEN:
                raise housecanary.exceptions.UnauthorizedException(code, message)
            else:
                raise housecanary.exceptions.RequestException(code, message)

        request_url = response.request.url

        endpoint_name = self._parse_endpoint_name_from_url(request_url)

        return Response.create(endpoint_name, response_json, response)

    def process_pdf_response(self, response):
        return response.text

    def process_zip_response(self, response):
        return response.text

    @staticmethod
    def _parse_endpoint_name_from_url(request_url):
        # get the path from the url
        path = urlparse(request_url).path

        # path is like "/v2/property/value"

        # strip off the leading "/"
        path = path[1:]

        # keep only the part after the version and "/"
        path = path[path.find("/")+1:]

        # path is now like "property/value"
        return path



import os
from housecanary.output import ResponseOutputGenerator
from housecanary.requestclient import RequestClient
import housecanary.exceptions
import housecanary.constants as constants
from requests.auth import HTTPBasicAuth


class ApiClient(object):
    """Base class for making API calls"""

    def __init__(self, auth_key=None, auth_secret=None, version=None, request_client=None,
                 output_generator=None, auth=None):
        """
        auth_key and auth_secret can be passed in as parameters or
        pulled automatically from the following environment variables:
            HC_API_KEY -- Your HouseCanary API auth key
            HC_API_SECRET -- Your HouseCanary API secret

        Passing in the key and secret as parameters will take precedence over environment variables.

        Args:
            auth_key - Optional. The HouseCanary API auth key.
            auth_secret - Optional. The HouseCanary API secret.
            version (str) - Optional. The API version to use, in the form "v2", for example.
                            Default is "v2".
            request_client - Optional. An instance of a class that is responsible for
                             making http requests. It must implement a `post` method.
                             Default is RequestClient.
            output_generator - Optional. An instance of an OutputGenerator that implements
                               a `process_response` method. Can be used to implement custom
                               serialization of the response data from the request client.
                               Default is ResponseOutputGenerator.
            authenticator - Optional. An instance of a requests.auth.AuthBase implementation
                            for providing authentication to the request.
                            Default is requests.auth.HTTPBasicAuth.
        """

        self._auth_key = auth_key or os.environ['HC_API_KEY']
        self._auth_secret = auth_secret or os.environ['HC_API_SECRET']

        self._version = version or constants.DEFAULT_VERSION

        # user can pass in a custom request_client
        self._request_client = request_client

        # if no request_client provided, use the defaults.
        if self._request_client is None:
            # allow using custom OutputGenerator or Authenticator with the RequestClient
            _output_generator = output_generator or ResponseOutputGenerator()
            _auth = auth or HTTPBasicAuth(self._auth_key, self._auth_secret)
            self._request_client = RequestClient(_output_generator, _auth)

        self.property = PropertyComponentWrapper(self)

    def fetch(self, endpoint_name, post_data, query_params=None):
        """Calls this instance's request_client's post method with the
        specified component endpoint

        Args:
            - endpoint_name (str) - The endpoint to call like "property/value".
            - address_data - One or more address to request data for. Address data can
                             be in one of these forms:

                            - A Json formatted list like:
                            [{"address":"82 County Line Rd",
                              "zipcode":"72173",
                              "meta":"extra info"}]

                                - A list of (address, zipcode, meta) tuples like:
                                [("82 County Line Rd", "72173", "extra info")]

                                - A single address tuple:
                                ("82 County Line Rd", "72173", "extra_info")

                                The "meta" field is optional.

        Returns:
            A Response object, or the output of a custom OutputGenerator
            if one was specified in the constructor.
        """

        if query_params is None:
            query_params = {}

        endpoint_url = constants.URL_PREFIX + "/" + self._version + "/" + endpoint_name
        return self._request_client.post(endpoint_url, post_data, query_params)

    def fetch_synchronous(self, endpoint_name, query_params=None):
        """Calls this instance's request_client's get method with the
        specified component endpoint"""

        endpoint_url = constants.URL_PREFIX + "/" + self._version + "/" + endpoint_name

        if query_params is None:
            query_params = {}

        return self._request_client.get(endpoint_url, query_params)


class PropertyComponentWrapper(object):
    """Property specific components

    All the component methods of this class take address data as parameter. Address data can
    be in one of two forms:

        - A Json formatted list like:
        [{"address":"82 County Line Rd", "zipcode":"72173", "meta":"extra info"}]

        - A list of (address, zipcode, meta) tuples like:
        [("82 County Line Rd", "72173", "extra info")]

        The "meta" field is optional.

    All the component methods of this class return a Response object,
    or the output of a custom OutputGenerator if one was specified in the constructor.
    """

    def __init__(self, api_client=None):
        """
        Args:
            - api_client - An instances of ApiClient
        """
        self._api_client = api_client

    def _fetch_property_component(self, endpoint_name, address_data, query_params=None):
        """common method for handling parameters before passing to api_client"""

        if query_params is None:
            query_params = {}

        post_data = get_post_data(address_data)

        return self._api_client.fetch(endpoint_name, post_data, query_params)

    def census(self, address_data):
        """Call the census endpoint"""
        return self._fetch_property_component("property/census", address_data)

    def details(self, address_data):
        """Call the details endpoint"""
        return self._fetch_property_component("property/details", address_data)

    def flood(self, address_data):
        """Call the flood endpoint"""
        return self._fetch_property_component("property/flood", address_data)

    def mortgage_lien(self, address_data):
        """Call the mortgage_lien endpoint"""
        return self._fetch_property_component("property/mortgage_lien", address_data)

    def msa_details(self, address_data):
        """Call the msa_details endpoint"""
        return self._fetch_property_component("property/msa_details", address_data)

    def nod(self, address_data):
        """Call the nod endpoint"""
        return self._fetch_property_component("property/nod", address_data)

    def owner_occupied(self, address_data):
        """Call the owner_occupied endpoint"""
        return self._fetch_property_component("property/owner_occupied", address_data)

    def rental_value(self, address_data):
        """Call the rental_value endpoint"""
        return self._fetch_property_component("property/rental_value", address_data)

    def sales_history(self, address_data):
        """Call the sales_history endpoint"""
        return self._fetch_property_component("property/sales_history", address_data)

    def school(self, address_data):
        """Call the school endpoint"""
        return self._fetch_property_component("property/school", address_data)

    def value(self, address_data):
        """Call the value endpoint"""
        return self._fetch_property_component("property/value", address_data)

    def value_forecast(self, address_data):
        """Call the value_forecast endpoint"""
        return self._fetch_property_component("property/value_forecast", address_data)

    def zip_details(self, address_data):
        """Call the zip_details endpoint"""
        return self._fetch_property_component("property/zip_details", address_data)

    def zip_hpi_forecast(self, address_data):
        """Call the zip_hpi_forecast endpoint"""
        return self._fetch_property_component("property/zip_hpi_forecast", address_data)

    def zip_hpi_historical(self, address_data):
        """Call the zip_hpi_historical endpoint"""
        return self._fetch_property_component("property/zip_hpi_historical", address_data)

    def zip_volatility(self, address_data):
        """Call the zip_volatility endpoint"""
        return self._fetch_property_component("property/zip_volatility", address_data)

    def component_mget(self, address_data, components):
        """Call the component_mget endpoint

        Args:
            - address_data - As described in the class docstring.
            - components - A list of strings for each component to include in the request.
                Example: ["property/details", "property/flood", "property/value"]
        """
        query_params = {"components": ",".join(components)}

        return self._fetch_property_component("property/component_mget", address_data, query_params)

    def value_report(self, address, zipcode, report_type="full", format_type="all"):
        """Call the value_report component

        Value Report only supports a single address.

        Args:
            - address
            - zipcode

        Kwargs:
            - report_type - "full" or "summary". Default is "full".
            - format_type - "json", "pdf" or "all". Default is "json".
        """
        query_params = {
            "report_type": report_type,
            "format": format_type,
            "address": address,
            "zipcode": zipcode
        }

        return self._api_client.fetch_synchronous("property/value_report", query_params)


def get_post_data(address_data):
    """Convert the various formats of input address_data into
    the proper json format expected by the API."""

    post_data = []

    if isinstance(address_data, list) and len(address_data) > 0:
        # if list, convert each address in the list to json
        for address in address_data:
            post_data.append(convert_to_json(address))
    else:
        post_data.append(convert_to_json(address_data))

    return post_data


def convert_to_json(address_data):
    """Convert input address tuples into json format"""

    if address_data and isinstance(address_data, str):
        # allow just passing an address string, for future support.
        # The api currently requires a zipcode as well, but may not in the future.
        address_json = {}
        address_json["address"] = address_data
        return address_json

    if isinstance(address_data, tuple) and len(address_data) > 0:
        address_json = {}
        address_json["address"] = address_data[0]
        if len(address_data) > 1:
            address_json["zipcode"] = address_data[1]
        if len(address_data) > 2:
            address_json["meta"] = address_data[2]
        return address_json

    elif isinstance(address_data, dict):
        allowed_keys = ["address", "zipcode", "meta"]

        # ensure the dict does not contain any unallowed keys
        for key in address_data:
            if key not in allowed_keys:
                msg = "Key in address input not allowed: " + key
                raise housecanary.exceptions.InvalidInputException(msg)

        # ensure it contains an "address" key
        if "address" in address_data:
            return address_data

    # if we made it here, the input was not valid.
    msg = ("Input is invalid. Must be a list of (address, zipcode) tuples,"
           " or a Json formatted list with each item containing at least an 'address' key.")
    raise housecanary.exceptions.InvalidInputException((msg))



from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='housecanary',
      version='0.3',
      description='Client Wrapper for the HouseCanary API',
      long_description=readme(),
      url='http://github.com/housecanary/hc-api-python',
      author='HouseCanary',
      author_email='techops@housecanary.com',
      license='MIT',
      packages=['housecanary'],
      install_requires=['requests'],
      zip_safe=False,
      test_suite='nose.collector',
      tests_require=['nose', 'mock'])